<?php
// 投稿 (Post) 関連の関数

// [投稿ファイルについて]
// 投稿ファイルは POST_DIR (data/post) フォルダーに保存される。
// 投稿ファイルの名前にメタ情報が含まれる。
// 形式:
//  "{$datetime}.id{$id}.us{$userid}.to{未使用}.th{$thread_id}.txt"
// 内訳:
//  $datetime: 投稿日時(ISO形式、タイムゾーンは「Z」固定)
//  $id: 投稿ID(md5ハッシュ)
//  $userid: 投稿者$id (ゲストの場合「-」が入る)
//  {未使用}: 現在は常に「-」が入る
//  $thread_id: 返信先スレッドの投稿ID (返信以外なら「-」が入る)

// ファイル名先頭の日附は、1日あたりの投稿数の判定や、
// 新着順のソートなどに使用できる。

// 投稿ファイルの内容はヘッダー情報、本文の2つのパートに分れ、
// 両者の間には「-」だけの行がある。
// ヘッダー情報のパートは「項目=値」の形式で屬性を持つ。
// 内訳:
//  title=件名
//  attachment_id=添付ファイルのID

// [添附ファイルについて]
// 添附ファイルは ATTACHMENT_DIR (data/attachment) フォルダーに gzip 形式で保存される。
/// ファイル名の形式: {id}.gz
// idにはランダムな英數字が入る。

// [聯投防止について]
// 投稿されてから一定時間は同じ内容の投稿ができなくなる。
// 制限の狀態は LIMIT_TSV (data/limit.tsv) に保存される。
// 制限は「投稿から記號を削除した後の文字列」で判定される。
// 制限される時間などは make_repeating_info を參照。

$post_title_cache = [];

$date_now = date('Y-m-d\\TH:i:s\\Z');

function load_postfile($filepath) {
	$text = file_get_contents($filepath);
	$split = preg_split('/^-\n/m', $text, 2);
	unset($text);

	$head = $split[0];
	$body = $split[1] ?? '';
	unset($split);

	$parsed = parse_key_value($head, ['title', 'attachment_id']);
	unset($head);

	$title = $parsed['title'] ?? '';
	$attachment_id = $parsed['attachment_id'] ?? '';
	unset($parsed);

	$deleted = $body === '';
	if ($deleted) {
		$body = DELETEDTEXT;
		$title = '';
		$attachment_id = '';
	}

	return compact('deleted', 'title', 'body', 'attachment_id');
}

function get_post_metadata($file, $deleted = false) {
	global $date_now;

	// Filename's head is datetime
	$datetime = substr($file, 0, 20);
	$is_future = strcmp($datetime, $date_now) > 0;

	$date = new Datetime($datetime);
	if (TIMEZONE !== null) {
		$date->setTimezone(new DateTimeZone(TIMEZONE));
	}
	$format = 'Y-m-d';
	if (!HIDE_TIME || $is_future) { $format .= ' H:i'; }
	$time = $date->format($format);
	unset($format, $date);

	// TODO:これだとファイル名がマッチしない時に予期せぬ結果になる

	// 投稿ID: ".id{$id}."
	$id = preg_replace('/.+\.id([^.]+)\..+/', '\1', $file);

	// 返信先スレッドID: ".th{$id}."
	$thread_id = preg_replace('/.+\.th([^.]+)\..+/', '\1', $file);
	if ($thread_id === NOREPLY_ID) { $thread_id = ''; }

	// 投稿ユーザーID: ".us{$id}"
	$userid = GUEST_ID;
	if (!$deleted) {
		$userid = preg_replace('/.+\.us([^.]+)\..+/', '\1', $file);
	}
	$is_guest = $userid === GUEST_ID;
	if ($is_guest) {
		$userid = '';
	}

	$thread_title = '';
	if ($thread_id != '') {
		$thread_title = load_post_title_by_id($thread_id);
	}

	$detail_url = sitebase('post/?id=' . $id);
	$delete_url = sitebase('post-delete/?id=' . $id);
	$user_url = $is_guest ? '' : sitebase('user/?id=' . $userid);
	$thread_url = $thread_id != '' ? sitebase('post/?id=' . $thread_id) : '';
	$is_mine = !$is_guest && $userid === ($_SESSION['user']['id'] ?? '');

	// too long, sorry.
	return compact('id', 'is_guest', 'userid', 'detail_url', 'delete_url', 'user_url', 'datetime', 'time', 'thread_id', 'thread_title', 'thread_url', 'is_mine', 'is_future');
}

function load_res_num($id) {
	$filepath = filepath_by_post_id($id);
	if ($filepath === null) { return 1; }

	$file = basename($filepath);
	$meta = get_post_metadata($file);
	$thread_id = $meta['thread_id'];
	unset($meta);
	if ($thread_id == '') { return 1; }

	$pattern = "2*Z*.th{$thread_id}.txt";
	$list = glob(POST_DIR . $pattern);
	$search = ".id{$id}.";
	foreach ($list as $i => $filepath) {
		if (strpos(basename($filepath), $search) > 0) {
			return $i + 2;
		}
	}

	return 1; // fallback
}

function load_post($filepath) {
	$files = null;

	$detail = load_postfile($filepath);
	$deleted = $detail['deleted'];
	$title = $detail['title'];
	$has_title = $title != '';
	$body_raw = $detail['body'];
	$attachment_id = $detail['attachment_id'];
	unset($detail);

	$metadata = get_post_metadata(basename($filepath), $deleted);
	if ($metadata['thread_id'] == '' && $title == '') {
		$title = post_default_title($metadata['id']);
	}
	$userid = $metadata['userid'];
	$users = load_users();

	$username = GUESTNAME;
	if ($deleted) {
		$username = DELETEDNAME;
	}
	elseif ($userid != '') {
		$username = $users[$userid]['username'] ?? 'UNKNOWN';
	}

	$body = url_to_link(
		nl2br(
			anchor_to_link(
				htmlspecialchars($body_raw),
				$metadata['thread_id'] ?? $metadata['id']
			),
			false
		)
	);

	$attachments = [];
	if ($attachment_id != '') {
		$attachments[] = [
			'url' => sitebase('attachment/?id=' . $attachment_id),
			'name' => '添付画像(Image)',
		];
	}

	return array_merge(
		$metadata,
		compact('username', 'deleted', 'title', 'has_title', 'body', 'body_raw', 'attachments'),
	);
}

function load_post_by_id($id) {
	$filepath = filepath_by_post_id($id);
	if ($filepath === null) { return null; }
	return load_post($filepath);
}

function search_post($options = []) {
	global $date_now;
	if (!file_exists(POST_DIR)) {
		die(POST_DIR.': ディレクトリは存在しません。');
	}

	$includes_future = $options['includes_future'] ?? false;
	$pagesize = $options['pagesize'] ?? 0;
	$has_paging = $pagesize > 0;
	$thread = false;

	$pager_prefix = $options['pager_prefix'] ?? '?page=';
	$key = $options['key'] ?? '';
	$value = $options['value'] ?? '';
	$page = $options['page'] ?? 1;
	unset($options);

	$pattern = '2*Z*.txt';
	if ($key === 'userid' && $value != '' && $value !== 'tl') {
		$pattern = "2*Z*.us{$value}.*.txt";
	}
	elseif ($key == 'thread' && $value >= 0) {
		// 'th' is last
		$pattern = "2*Z*.th{$value}.txt";
		$thread = true;
	}
	unset($key, $value);

	$files = [];
	$files_wk = glob(POST_DIR . $pattern);
	$len = sizeof($files_wk);
	for ($i = 0; $i < $len; $i++) {
		$idx = $thread ? $i : $len - $i - 1;
		if (!$includes_future && strcmp(basename($files_wk[$idx]), $date_now) > 0) {

			continue;
		}
		$files[] = $files_wk[$idx];
	}
	unset($files_wk, $i, $len);

	$total = sizeof($files);
	if ($has_paging) {
		$last_page = (int)ceil($total / $pagesize);
	}
	else {
		$last_page = 0;
	}
	$page = max(1, min($last_page, $page));

	if ($has_paging) {
		$tail = array_slice($files, ($page - 1) * $pagesize, $pagesize);
	}
	else {
		$tail = $files;
	}
	unset($files);

	$post_list = [];
	foreach ($tail as $filepath) {
		$post_list[] = load_post($filepath);
	}
	unset($filepath, $tail);

	$has_pager = $last_page > 1;
	$pager = [];
	if ($has_paging && $has_pager) {
		if ($page > 1) {
			$pager['after'] = $pager_prefix . ($page - 1);
		}
		if ($page < $last_page) {
			$pager['before'] = $pager_prefix . ($page + 1);
		}
		$pager['pagesize'] = POSTS_PER_PAGE;
	}

	return compact('post_list', 'total', 'page', 'pager', 'last_page');
}

function add_post($userid, $title, $body, $attachment_id, $file_hash, &$post_id_ref, $thread_id, $spooftime) {
	// TODO:引数は配列の方が良い?
	// TODO:$post_id_refの参照渡しはイマイチ
	if ($thread_id !== NOREPLY_ID) {
		// 返信の場合件名を捨てる
		$title = '';
	}

	$text = '';
	if ($title) {
		$text .= 'title=' . $title . PHP_EOL;
	}
	if (isset($attachment_id)) {
		$text .= 'attachment_id=' . $attachment_id . PHP_EOL;
	}
	$text .= '-' . PHP_EOL . $body . PHP_EOL;

	$id = md5(bin2hex(random_bytes(256)) . $userid . $title . $body);

	$dto = new DateTime($spooftime ? '+' . (mt_rand(60 * 60 * 3, 60 * 60 * 27 - 1)) . 'seconds' : '');
	$datetime = $dto->format('Y-m-d\\TH:i:s\\Z');
	$timestamp = $dto->getTimestamp();
	unset($dto);
	$file = "{$datetime}.id{$id}.us{$userid}.to-.th{$thread_id}.txt";

	if ($body !== '.' || $file_hash === '') {
		$error = check_repeating_info($body, $datetime);
		if ($error) { return ['すでに同様の書き込み有り(削除済みを含む)。連投防止。']; }
	}
	if ($file_hash != '') {
		$error = check_repeating_info($body, $datetime, $file_hash);
		if ($error) { return ['すでに同じ画像有り(削除済みを含む)。連投防止。']; }
	}

	if (!mkdir_p(POST_DIR)) {
		die('mkdir_pを実行に失敗しました。');
	}

	$result = file_put_contents(POST_DIR . $file, $text, LOCK_EX);
	if ($result === false) {
		die('書き込みの保存に失敗しました。');
	}

	touch(POST_DIR . $file, $timestamp, $timestamp);
	chmod(POST_DIR . $file, 0644);

	if ($body !== '.' || $file_hash === '') {
		add_repeating_info(make_repeating_info($body, $datetime));
	}
	if ($file_hash != '') {
		add_repeating_info(make_repeating_info($body, $datetime, $file_hash));
	}

	$post_id_ref = $id;
	return [];
}

function check_uploaded_image($key) {
	$size = $_FILES[$key]['size'] ?? 0;
	if ($size <= 0) { return ['画像ファイルが不正。']; }
	else if ($size > 1024 * 500) { ['画像ファイルは 500 kb 以内。']; }

	$buffer = file_get_contents($_FILES[$key]['tmp_name']);
	$type = get_image_type($buffer);
	unset($buffer);
	if (!isset($type)) { return ['画像ファイルが不正。']; }

	return [];
}

function save_uploaded_image($key, $attachment_id) {
	$buffer = file_get_contents($_FILES[$key]['tmp_name']);
	$type = get_image_type($buffer);
	if (!isset($type)) { on_error(400, ['画像ファイルが不正。']); }

	if (!$result = mkdir_p(ATTACHMENT_DIR, 0700)) {
		return ['フォルダの作成に失敗。'];
	}

	$target = ATTACHMENT_DIR . $attachment_id . '.gz';
	if (!$gz = gzopen($target, 'w1')) {
		die('画像の書き込みに失敗しました。');
	}

	$result = gzwrite($gz, $buffer);
	if ($result === false) {
		die('画像の書き込みに失敗しました。');
	}

	gzclose($gz);
	chmod($target, 0777);

	return [];
}

function delete_post($id, $hard = false) {
	$files = glob(POST_DIR . '2*Z*.id' . $id . '.*.txt');
	if (sizeof($files) !== 1) { return ['書き込みのファイルが存在しない。']; }

	$filepath = $files[0];
	unset($files);

	$attachment_id = load_postfile($filepath)['attachment_id'];

	// 存在確認済みだからmkdir_pは不要

	if ($attachment_id) {
		$attachment_path = ATTACHMENT_DIR . $attachment_id . '.gz';
		if (file_exists($attachment_path)) {
			if (!unlink($attachment_path)) {
				die('ファイルの削除に失敗。');
			}
		}
	}

	if ($hard) {
		$result = unlink($filepath);
	} else {
		$result = file_put_contents($filepath, '', LOCK_EX);
	}
	if ($result === false) {
		return die('書き込みの削除に失敗。');
	}

	return [];
}

function load_threads() {
	$pattern = '2*Z.*.th*.txt';
	$files = glob(POST_DIR . $pattern);
	$threads = [];

	$i = sizeof($files);
	while ($i > 0) {
		$i--;

		$filepath = $files[$i];
		$metadata = get_post_metadata(basename($filepath));
		$is_future = $metadata['is_future'];
		$thread_id = $metadata['thread_id'];
		unset($metadata);

		if ($is_future) { continue; }
		if ($thread_id === '') { continue; }

		if (isset($threads[$thread_id])) {
			$threads[$thread_id]['count'] += 1;
			continue;
		}

		$parent = load_post_by_id($thread_id);
		if ($parent['deleted']) { continue; }

		$text = preg_replace("/[\n\r]+/", ' ', $parent['body_raw']);
		$thread = [
			'count' => 2,
			'title' => $parent['title'] ?? '',
			'detail_url'=> $parent['detail_url'],
			'sammary' => htmlspecialchars(mb_substr($text, 0, 30)),
		];
		if (mb_strlen($text) > 30) {
			$thread['sammary'] .= '(...)';
		}
		unset($text);

		$threads[$thread_id] = $thread;
		unset($thread);
	}
	unset($files, $filepath);

	return $threads;
}

function make_repeating_info($body, $datetime, $file_hash = '') {
	if ($file_hash != '') {
		return ['hash' => $file_fash, 'limit' => mb_substr($datetime, 0, 10)];
	}

	$letter_only = preg_replace('/[^\p{Ll}\p{Lt}\p{Lo}\p{Lu}\p{N}]/u', '', $body);
	if (mb_strlen($letter_only) > 0) { $body = $letter_only; }

	$hash = md5($body);

	$len = mb_strlen($body);
	if ($len <= 2) { $size = 16; }
	elseif ($len <= 30) { $size = 13; }
	elseif ($len <= 40) { $size = 10; }
	else { $size = 7; }
	$limit = mb_substr($datetime, 0, $size);

	return compact('hash', 'limit');
}

function add_repeating_info($repeating_info) {
	// NOTE:大して重要ではないためロックしない
	$hash = $repeating_info['hash'] ?? '';
	$limit = $repeating_info['limit'] ?? '';
	$text = "$hash\t$limit" . PHP_EOL;
	$result = file_put_contents(LIMIT_TSV, $text, FILE_APPEND | LOCK_EX);
}

function check_repeating_info($body, $datetime, $file_hash = '') {
	if (!file_exists(LIMIT_TSV)) {
		touch(LIMIT_TSV, time());
		chmod(LIMIT_TSV, 0644);
	}
	$hash = make_repeating_info($body, $datetime, $file_hash)['hash'];
	$text = file_get_contents(LIMIT_TSV) ?? '';
	$lines = explode(PHP_EOL, $text);
	unset($text);

	$out = '';
	$count = 0;

	foreach ($lines as $line) {
		if (mbtrim($line) === '') { continue; }

		$array = explode("\t", $line);
		if (strpos($datetime, $array[1] ?? '') === 0) {
			$out .= $line . PHP_EOL;
			if ($array[0] === $hash) {
				$count += 1;
			}
		}

		unset($array);
		unset($line);
	}

	unset($lines);

	$result = file_put_contents(LIMIT_TSV, $out, LOCK_EX);

	return $count >= 1;
}

// Util

function count_post() {
	if (!file_exists(POST_DIR)) {
		die(POST_DIR.': ディレクトリは存在しません。');
	}

	$pattern = '2*Z*.txt';
	$files = glob(POST_DIR . '2*Z*.txt');
	$size = sizeof($files);
	unset($files);

	return $size;
}

function count_post_today() {
	if (!file_exists(POST_DIR)) {
		die(POST_DIR.': ディレクトリは存在しません。');
	}

	$date = date('Y-m-d');
	$pattern = "{$date}T*Z.*.txt";
	$files = glob(POST_DIR . $pattern);
	$size = sizeof($files);
	unset($files);

	return $size;
}

function filepath_by_post_id($id) {
	if (!file_exists(POST_DIR)) {
		die(POST_DIR.': ディレクトリは存在しません。');
	}

	$files = glob(POST_DIR . '2*Z*.id' . $id . '.*.txt');
	if (sizeof($files) !== 1) {
		return null;
	}
	return $files[0];
}

function load_post_title_by_id($id) {
	if (!file_exists(POST_DIR)) {
		die(POST_DIR.': ディレクトリは存在しません。');
	}

	global $post_title_cache;

	if (isset($post_title_cache[$id])) { return $post_title_cache[$id]; }

	$files = glob(POST_DIR . '2*Z*.id' . $id . '.*.txt');
	if (sizeof($files) != 1) {
		unset($files);
		$post_title_cache[$id] = '';
		return '';
	}

	$filepath = $files[0];
	unset($files);
	$title = load_postfile($filepath)['title'] ?? '';
	if ($title === '') {
		$title = post_default_title($id);
	}
	$post_title_cache[$id] = $title;

	return $title;
}
